import os.path
import unittest
import tempfile
from dekgen.utils.yaml import yaml
from dekgen.utils.yaml.tags import tmpl_data_final
from dekgen.tmpl.render import render_dir
from dektools.output import pprint
from dektools.file import read_text


class MyTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_yaml(self):
        pprint(tmpl_data_final(yaml.load('./res/test.yaml')))

    def test_render(self):
        target = tempfile.mkdtemp()
        render_dir(target, './render', files=['./render/values.test.yaml'])
        print(read_text(os.path.join(target, 'main.txt')))



from pathlib import Path
from functools import partial, lru_cache
from .template import Template


def tab_str(s, n, p=4, sl=False):  # s: list of str or str
    if isinstance(s, str):
        s = [s]
    r = []
    for x in s:
        if sl:
            x = x.strip()
            if x:
                r.append(x)
        else:
            r.append(x)
    r = '\n'.join(r).split('\n')
    return '\n'.join([' ' * n * p + x for x in r])


class Cache:
    def __init__(self, master):
        self.master = master
        self.caches = {}

    def __getitem__(self, method):
        return partial(self.__run, method)

    def __getattr__(self, method):
        return self[method]

    def __run(self, method):
        if method not in self.caches:
            self.caches[method] = getattr(self.master, method)()
        return self.caches[method]


class Repr:
    def __init__(self, value, delay=False):
        self.value = value
        self.delay = delay

    def final_value(self):
        return self.value

    def __repr__(self):
        value = self.final_value()
        return f'lambda: {value}' if self.delay else value

    def __copy__(self):
        return self.__class__(self.value, self.delay)

    def __deepcopy__(self, memo):
        return self.__class__(self.value, self.delay)


class ReprTuple(Repr):
    def __init__(self, value, delay=False):
        super().__init__(list(value), delay)

    def final_value(self):
        return repr(tuple(self.value))

    def __getattr__(self, item):
        return getattr(self.value, item)


class Generator:
    TEMPLATE_DIR = None

    template_name = None
    template_cls = Template
    template_ext = '.tpl'

    Repr = Repr
    ReprTuple = ReprTuple

    env_default = {}

    def __init__(self, target_dir, instance, kwargs=None):
        self.target_dir = str(target_dir)
        self.instance = instance
        self.kwargs = kwargs or {}
        self.cache = Cache(self)

    def check(self):
        return bool(self.instance)

    def action(self):
        if self.check():
            self.render()
            self.on_rendered()

    def on_rendered(self):
        pass

    def render(self):
        self.template_cls(self.normalize_variables(
            {**self.variables, **self.kwargs}), self.env_default).render_dir(self.target_dir, self.template_path)

    def render_tpl(self, tpl, variables=None):
        return self.template_cls(self.normalize_variables(
            {**(self.variables if variables is None else variables), **self.kwargs}), self.env_default).render_string(
            str(Path(self.template_path) / (tpl + self.template_ext)))

    def normalize_variables(self, variables):
        return variables

    @property
    @lru_cache(None)
    def template_path(self):
        return str(Path(self.TEMPLATE_DIR) / self.template_name) if self.template_name else str(self.TEMPLATE_DIR)

    @property
    @lru_cache(None)
    def variables(self):
        return self.collect_data('variables')

    def get_variables(self, *args, **kwargs):
        return {
            **{k: self.variables[k] for k in args},
            **{k: self.variables[v] for k, v in kwargs.items()}
        }

    def collect_data(self, name, collection=dict):
        result = collection()
        pre_collect = f'pre_collect_{name}'
        post_collect = f"post_collect_{name}"
        if hasattr(self, pre_collect):
            result = collection(getattr(self, pre_collect)() or collection())
        for x in dir(self):
            if x.startswith(f'{name}_'):
                v = getattr(self, x)
                result.update(collection(self.cache[x]() if callable(v) else v))
        if hasattr(self, post_collect):
            result = collection(getattr(self, post_collect)(result))
        return result

    @staticmethod
    def tab_str(s, n, p=4, sl=False):  # s: list or str
        return tab_str(s, n, p, sl)



import os
from jinja2 import Environment
from dektools.file import read_text, read_file, write_file, list_relative_path, FileHitChecker, normal_path


class Template:
    template_suffix = '-tpl'
    target_suffix = None

    file_ignore_tpl = ['.ignoretpl', '.ignoretpl.tpl']
    file_ignore_override = ['.ignoreor', '.ignoreor.tpl']
    file_ignore = ['.ignore', '.ignore.tpl']

    default_env_kwargs = dict(
        autoescape=False
    )

    def __init__(self, variables, env=None, **kwargs):
        self.args_copy = [variables, kwargs]
        self.env = Environment(**{
            **self.default_env_kwargs,
            **kwargs
        })
        self.variables = variables
        if env:
            for k, data in env.items():
                getattr(self.env, k).update(data)

    def copy(self, variables=None):
        return self.__class__(
            {**self.args_copy[0], **(variables or {})},
            **self.args_copy[1]
        )

    def render(self, content):
        return self.env.from_string(content).render(self.variables)

    def render_string(self, template_file, close_tpl=False):
        if close_tpl:
            return read_file(template_file)
        else:
            return self.render(read_text(template_file))

    def render_file(self, template_file, target_file, close_tpl=False):
        content = self.render_string(template_file, close_tpl)
        write_file(target_file, sb=content)
        return content

    def render_dir(self, target_dir, template_path, force_close_tpl=False, open_ignore_override=True):
        ignore_tpl = FileHitChecker(template_path, self.file_ignore_tpl[0],
                                    rules=self.get_hit_rules(template_path, self.file_ignore_tpl[1]))
        ignore_override = FileHitChecker(template_path, self.file_ignore_override[0],
                                         rules=self.get_hit_rules(template_path, self.file_ignore_override[1]))
        ignore = FileHitChecker(template_path, self.file_ignore[0],
                                rules=[*self.get_hit_rules(template_path, self.file_ignore[1]),
                                       *[f'/{item}' for item in
                                         [*self.file_ignore_tpl, *self.file_ignore_override, *self.file_ignore]]])
        path_virtual = os.path.join(template_path, '__virtual__')
        variables_virtual = {}
        for rp, fp in list_relative_path(path_virtual).items():
            variables_virtual[rp.replace('\\', '/').replace('/', '__')] = self.render_string(fp)
        self.variables['__virtual__'] = variables_virtual
        for root, _, files in os.walk(template_path):
            for f in files:
                fp = os.path.join(root, f)
                if ignore.is_hit(fp):
                    continue
                fn, ext = splitext(fp)
                ext = ext.rsplit(self.template_suffix, 1)[0] if self.template_suffix else ext
                rp = (fn + ext)[len(template_path):]
                target_file = target_dir + self.render(rp)
                if open_ignore_override and os.path.exists(target_file) and ignore_override.is_hit(fp):
                    continue
                self.render_file(fp, target_file, force_close_tpl or ignore_tpl.is_hit(fp))

    def get_hit_rules(self, template_path, filepath):
        ignore_file = normal_path(os.path.join(template_path, filepath))
        if os.path.exists(ignore_file):
            content = self.render_string(ignore_file)
            return [line for line in content.split('\n') if line]
        else:
            return []


class TemplateWide(Template):
    default_env_kwargs = dict(
        **Template.default_env_kwargs,
        variable_start_string='(=(',
        variable_end_string=')=)'
    )


def splitext(s):
    lst = s.rsplit('/', 1)
    if len(lst) == 2:
        dp, fn = lst
        sep = '/'
    else:
        lst = s.rsplit('\\', 1)
        sep = '\\'
        if len(lst) == 2:
            dp, fn = lst
        else:
            dp, fn = None, lst[0]
    lst = fn.rsplit('.', 1)
    if len(lst) == 2:
        fp, ext = lst
    else:
        fp, ext = lst[0], None
    ext = '' if ext is None else f'.{ext}'
    if dp is None:
        return fp, ext
    else:
        return dp + sep + fp, ext



import functools
from .imports import Imports, ImportsStatement
from .template import TemplateFormatted
from .template.env import env_extend
from ...tmpl.generator import Generator, Repr


class ReprTS(Repr):
    def final_value(self):
        return f'_({repr(self.value)})'


class CodeGenerator(Generator):
    ReprTS = ReprTS
    template_cls = TemplateFormatted
    _imports_cls = Imports
    _imports_statement_cls = ImportsStatement

    env_default = {**env_extend.context}

    template_codes_body = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._imports_statement = None
        self.parent = None

    def get_codes_body(self):
        if self.template_codes_body:
            return self.render_tpl(self.template_codes_body)
        else:
            return ''

    @property
    @functools.lru_cache(None)
    def iss(self):
        if self.parent_root:
            return self.parent_root._imports_statement
        else:
            _imports_core = getattr(self.__class__, '_imports_core', None)
            if _imports_core is None:
                _imports_core = self.__class__._imports_core = self._imports_cls()
            imports_core = _imports_core.derive()
            imports_core.update(self.walk_imports(self))
            self._imports_statement = self._imports_statement_cls(imports_core)
            return self._imports_statement

    @property
    @functools.lru_cache(None)
    def parent_root(self):
        cursor = self.parent
        while cursor:
            if cursor.parent:
                cursor = cursor.parent
            else:
                return cursor
        return cursor

    def post_collect_children(self, children):
        for lst in children.values():
            for child in lst:
                child.parent = self
        return children

    def post_collect_variables(self, variables):
        children_data = {}
        all_filter_children = getattr(self, f'all_filter_children', lambda k, x, y: x)
        for key, array in self.children.items():
            r = self.tab_str([node.get_codes_body() for node in array], 0)
            filter_children = getattr(self, f'filter_children_{key}', None)
            if filter_children:
                children_data[key] = filter_children(r, array)
            else:
                children_data[key] = all_filter_children(key, r, array)

        if self.parent_root:
            imports_data = {}
        else:
            imports_data = dict(
                imports=self.iss.as_statements()
            )

        return self.post_post_collect_variables({
            **variables,
            **children_data,
            **imports_data
        })

    def pre_pre_collect_variables(self):
        return {}

    def post_post_collect_variables(self, variables):
        return variables

    @property
    @functools.lru_cache(None)
    def imports(self):
        return self.collect_data('imports', dict)

    @classmethod
    def walk_imports(cls, node):
        imports = {}
        imports.update(node.imports)
        for children in node.children.values():
            for child in children:
                imports.update(cls.walk_imports(child))
        return imports

    @property
    @functools.lru_cache(None)
    def children(self):
        return self.collect_data('children', dict)



import os
from dektools.yaml import yaml
from dektools.dict import dict_merge
from dektools.attr import object_path_update
from .template import TemplateWide


class RenderTemplate(TemplateWide):
    file_ignore_tpl = [f"../{x}" for x in TemplateWide.file_ignore_tpl]
    file_ignore_override = [f"../{x}" for x in TemplateWide.file_ignore_override]
    file_ignore = [f"../{x}" for x in TemplateWide.file_ignore]


def render_dir(dest, src, files=None, updated=None, **kwargs):
    data = {}
    path_values = os.path.join(src, 'values.yaml')
    if os.path.isfile(path_values):
        data = dict(Values=yaml.load(path_values))
    if files:
        for f in files:
            strict = True
            if f.startswith('?'):
                f = f[1:]
                strict = False
            if not strict and not os.path.isfile(f):
                continue
            dict_merge(data, dict(Values=yaml.load(f)))
    if updated:
        object_path_update(data, updated)
    RenderTemplate(data, **kwargs).render_dir(dest, os.path.join(src, 'templates'))
